; ****************************************************************************
;                                 
;                                  Timer1
;
; ****************************************************************************

#include <avr/io.h>

.extern	Timer1Pwm
.extern Timer1Add
.extern Timer1Gen
.extern SinTab
.extern Timer1Seed
.extern Timer1His1
.extern Timer1His2
.extern Timer1His3
.extern Timer1His4
.extern Timer1His5
.extern Timer1His6
.extern Timer1SelNow
.extern MulDD

;#define USERND32	// flag - use 32-bit random to generate noise

	.text

; Timer1 overflow interrupt
;   Sine: 4+2+13+8+12+8+4+19+1+17 = 88 clocks
;   Triangle: 4+2+13+8+12+8+5+2+11+1+17 = 83 clocks
;   Sawtooth: 4+2+13+8+12+8+5+3+3+1+17 = 76 clocks
; 32-bit:
;   White Noise: 4+2+13+8+12+8+5+3+2+117+4+1+17 = 196 clocks
;   Pink Noise: 4+2+13+8+12+8+5+3+2+117+5+30+1+17 = 226 clocks
;   Brown Noise: 4+2+13+8+12+8+5+3+2+117+5+29+27+1+17 = 254 clocks
; 16-bit:
;   White Noise: 4+2+13+8+12+8+5+3+2+59+4+1+17 = 138 clocks
;   Pink Noise: 4+2+13+8+12+8+5+3+2+59+5+30+1+17 = 168 clocks
;   Brown Noise: 4+2+13+8+12+8+5+3+2+59+5+29+27+1+17 = 195 clocks

.global TIMER1_OVF_vect
TIMER1_OVF_vect:

	; interrupt [4]
	; rjmp [2]

; ----- push registers [13]

	push	r0			; [2]
	in	r0,_SFR_IO_ADDR(SREG)	; [1]
	push	r0			; [2]
	push	r31			; [2]
	push	r30			; [2]
	push	r29			; [2]
	push	r28			; [2]

; ----- load PWM counter [8]

	lds	r28,Timer1Pwm		; [2]
	lds	r29,Timer1Pwm+1		; [2]
	lds	r30,Timer1Pwm+2		; [2]
	lds	r31,Timer1Pwm+3		; [2]

; ----- add increment [12]

	lds	r0,Timer1Add		; [2]
	add	r28,r0			; [1]
	lds	r0,Timer1Add+1		; [2]
	adc	r29,r0			; [1]
	lds	r0,Timer1Add+2		; [2]
	adc	r30,r0			; [1]
	lds	r0,Timer1Add+3		; [2]
	adc	r31,r0			; [1]

; ----- save new PWM counter [8]

	sts	Timer1Pwm,r28		; [2]
	sts	Timer1Pwm+1,r29		; [2]
	sts	Timer1Pwm+2,r30		; [2]
	sts	Timer1Pwm+3,r31		; [2]

; ----- check SINE mode "3" [4,5]

	lds	r28,Timer1Gen		; [2]
	cpi	r28,3			; [1]
	brne	TIMER1_OVF_vect2	; [1,2]

; ----- load sine wave [19]

	lsr	r31			; [1]
	ror	r30			; [1]
	lsr	r31			; [1]
	ror	r30			; [1]
	lsr	r31			; [1]
	ror	r30			; [1]
	lsr	r31			; [1]
	ror	r30			; [1]
	lsr	r31			; [1]
	ror	r30			; [1]
	lsr	r31			; [1]
	ror	r30			; [1]

	subi	r30,lo8(-(SinTab))	; [1]
	sbci	r31,hi8(-(SinTab))	; [1]
	lpm	r31,Z			; [3]
	rjmp	TIMER1_OVF_vect6	; [2]

; ----- check TRIANGLE mode "4" [2,3]

TIMER1_OVF_vect2:
	cpi	r28,4			; [1]
	brne	TIMER1_OVF_vect4	; [1,2]

; ----- triangle [11]

	sbrs	r31,7			; [1,2,3]
	rjmp	TIMER1_OVF_vect3	; [2]
	com	r31			; [1]
	neg	r30			; [1]
	sbci	r31,0xff		; [1]
TIMER1_OVF_vect3:
	add	r30,r30			; [1]
	adc	r31,r31			; [1]
	rjmp	TIMER1_OVF_vect6	; [2]

; ----- check NOISE mode "6" [2,3]

TIMER1_OVF_vect4:
	cpi	r28,6			; [1]
	brne	TIMER1_OVF_vect6	; [1,2]

#ifdef USERND32	// flag - use 32-bit random to generate noise

; ----- noise 32-bit (random generator Timer1Seed = Timer1Seed*214013 + 2531011) [117]

	push	r1			; [2]
	push	r18			; [2]
	push	r19			; [2]
	push	r20			; [2]
	push	r21			; [2]
	push	r22			; [2]
	push	r23			; [2]
	push	r24			; [2]
	push	r25			; [2]
	push	r26			; [2]
	push	r27			; [2]

	lds	r22,Timer1Seed		; [2]
	lds	r23,Timer1Seed+1	; [2]
	lds	r24,Timer1Seed+2	; [2]
	lds	r25,Timer1Seed+3	; [2]

	ldi	r18,0xfd		; [1]
	ldi	r19,0x43		; [1]
	ldi	r20,0x03		; [1]
	ldi	r21,0x00		; [1]

	rcall	MulDD			; [48]

	subi	r22,0x3d		; [1]
	sbci	r23,0x61		; [1]
	sbci	r24,0xd9		; [1]
	sbci	r25,0xff		; [1]

	sts	Timer1Seed,r22		; [2]
	sts	Timer1Seed+1,r23	; [2]
	sts	Timer1Seed+2,r24	; [2]
	sts	Timer1Seed+3,r25	; [2]

	mov	r31,r25			; [1]

	pop	r27			; [2]
	pop	r26			; [2]
	pop	r25			; [2]
	pop	r24			; [2]
	pop	r23			; [2]
	pop	r22			; [2]
	pop	r21			; [2]
	pop	r20			; [2]
	pop	r19			; [2]
	pop	r18			; [2]
	pop	r1			; [2]

#else // USERND32

; ----- noise 16-bit (random generator Timer1Seed = Timer1Seed*60261 + 12403) [59]

	push	r1			; [2]
	push	r18			; [2]
	push	r19			; [2]
	push	r22			; [2]
	push	r23			; [2]
	push	r24			; [2]
	push	r25			; [2]

	lds	r24,Timer1Seed		; [2]
	lds	r25,Timer1Seed+1	; [2]

	ldi	r22,0x65		; [1]
	ldi	r23,0xeb		; [1]

	rcall	MulWW			; [18]

	subi	r24,0x8d		; [1]
	sbci	r25,0xcf		; [1]

	sts	Timer1Seed,r24		; [2]
	sts	Timer1Seed+1,r25	; [2]

	mov	r31,r25			; [1]

	pop	r25			; [2]
	pop	r24			; [2]
	pop	r23			; [2]
	pop	r22			; [2]
	pop	r19			; [2]
	pop	r18			; [2]
	pop	r1			; [2]

#endif // USERND32

; ----- 2=white noise, 1=pink noise, 0=brown noise [4,5]

	lds	r30,Timer1SelNow	; [2]
	cpi	r30,2			; [1]
	brne	TIMER1_OVF_vect8	; [1,2]

; ----- output OCR1A register [1]

TIMER1_OVF_vect6:
	out	_SFR_IO_ADDR(OCR1AL),r31 ; [1]

; ----- pop registers [17]

	pop	r28			; [2]
	pop	r29			; [2]
	pop	r30			; [2]
	pop	r31			; [2]
	pop	r0			; [2]
	out	_SFR_IO_ADDR(SREG),r0	; [1]
	pop	r0			; [2]
	reti				; [4]

; ----- pink or brown noise [29,30]

TIMER1_OVF_vect8:
	lds	r30,Timer1His1		; [2]
	lds	r29,Timer1His2		; [2]
	lds	r28,Timer1His3		; [2]

	sts	Timer1His1,r31		; [2]
	sts	Timer1His2,r30		; [2]
	sts	Timer1His3,r29		; [2]
	
	eor	r0,r0			; [1]
	add	r31,r30			; [1]
	adc	r0,r0			; [1]
	eor	r30,r30			; [1]
	add	r31,r29			; [1]
	adc	r30,r0			; [1]
	eor	r29,r29			; [1]
	add	r31,r28			; [1]
	adc	r30,r29			; [1]

	lsr	r30			; [1]
	ror	r31			; [1]
	lsr	r30			; [1]
	ror	r31			; [1]

	lds	r30,Timer1SelNow	; [2]
	cpi	r30,1			; [1]
	breq	TIMER1_OVF_vect6	; [1,2]

; ----- brown noise [27]

	lds	r30,Timer1His4		; [2]
	lds	r29,Timer1His5		; [2]
	lds	r28,Timer1His6		; [2]

	sts	Timer1His4,r31		; [2]
	sts	Timer1His5,r30		; [2]
	sts	Timer1His6,r29		; [2]
	
	eor	r0,r0			; [1]
	add	r31,r30			; [1]
	adc	r0,r0			; [1]
	eor	r30,r30			; [1]
	add	r31,r29			; [1]
	adc	r30,r0			; [1]
	eor	r29,r29			; [1]
	add	r31,r28			; [1]
	adc	r30,r29			; [1]

	lsr	r30			; [1]
	ror	r31			; [1]
	lsr	r30			; [1]
	ror	r31			; [1]
	rjmp	TIMER1_OVF_vect6	; [2]
