; ****************************************************************************
;
;                                   Division
;
; ****************************************************************************

;#define USE_DIVBB	; flag - use DivBB
;#define USE_DIVWW	; flag - use DivWW
;#define USE_DIVDD	; flag - use DivDD
#define USE_DIVQQ	; flag - use DivQQ

	.text

; ----------------------------------------------------------------------------
;                   Divide unsigned BYTE by BYTE, result is BYTE
; ----------------------------------------------------------------------------
; INPUT: R24 = dividend (N0)
;        R22 = divisor (D0)
;        R21:R20 = pointer to BYTE to store remainder (NULL=no remainder)
; OUTPUT: R24 = quotient (Q0)
; DESTROYS: R31, R30, R25
; takes 5.8 us on 16 MHz, 34 bytes
; ----------------------------------------------------------------------------

#ifdef USE_DIVBB

.global DivBB

; ----- clear dividend HIGH R25

DivBB:	sub	r25,r25

; ----- number of loops

	ldi	r31,8+1
	rjmp	DivBB4

; ----- shift dividend HIGH left (on input: carry from dividend LOW)

DivBB2:	adc	r25,r25

; ----- compare dividend HIGH with divisor

	cp	r25,r22
	brcs	DivBB4		; skip if dividend HIGH < divisor

; ----- subtract divisor from dividend HIGH (only in case if dividend HIGH >= divisor)

	sub	r25,r22

; ----- shift dividend LOW left (add Carry on input if dividend HIGH < divisor)

DivBB4:	adc	r24,r24

; ----- next loop

	dec	r31
	brne	DivBB2

; ----- invert quotient bits

	com	r24

; ----- store remainder

	cp	r20,r1
	cpc	r21,r1
	breq	DivBB6

	movw	r30,r20
	st	Z,r25

DivBB6:	ret

#endif

; ----------------------------------------------------------------------------
;                   Divide unsigned WORD by WORD, result is WORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24 = dividend (N1:N0)
;        R23:R22 = divisor (D1:D0)
;        R21:R20 = pointer to WORD to store remainder (NULL=no remainder)
; OUTPUT: R25:R24 = quotient (Q1:Q0)
; DESTROYS: R31, R30, R27, r26
; takes 13 us on 16 MHz, 48 bytes
; ----------------------------------------------------------------------------

#ifdef USE_DIVWW

.global DivWW

; ----- clear dividend HIGH R27:R26

DivWW:	sub	r26,r26
	sub	r27,r27

; ----- number of loops

	ldi	r31,16+1
	rjmp	DivWW4

; ----- shift dividend HIGH left (on input: carry from dividend LOW)

DivWW2:	adc	r26,r26
	adc	r27,r27

; ----- compare dividend HIGH with divisor

	cp	r26,r22
	cpc	r27,r23
	brcs	DivWW4		; skip if dividend HIGH < divisor

; ----- subtract divisor from dividend HIGH (only in case if dividend HIGH >= divisor)

	sub	r26,r22
	sbc	r27,r23

; ----- shift dividend LOW left (add Carry on input if dividend HIGH < divisor)

DivWW4:	adc	r24,r24
	adc	r25,r25

; ----- next loop

	dec	r31
	brne	DivWW2

; ----- invert quotient bits

	com	r24
	com	r25

; ----- store remainder

	cp	r20,r1
	cpc	r21,r1
	breq	DivWW6

	movw	r30,r20
	st	Z,r26
	std	Z+1,r27

DivWW6:	ret

#endif

; ----------------------------------------------------------------------------
;                   Divide unsigned DWORD by DWORD, result is DWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22 = dividend (N3:N2:N1:N0)
;        R21:R20:R19:R18 = divisor (D3:D2:D1:D0)
;        R17:R16 = pointer to DWORD to store remainder (NULL=no remainder)
; OUTPUT: R25:R24:R23:R22 = quotient (Q3:Q2:Q1:Q0)
;         R1=0
; DESTROYS: R31, R30, R27, R26
; takes 38 us on 16 MHz, 80 bytes
; ----------------------------------------------------------------------------

#ifdef USE_DIVDD

.global DivDD

; ----- number of loops

DivDD:	ldi	r26,32+1
	mov	r1,r26

; ----- clear dividend HIGH R31:R30:R27:R26

	sub	r26,r26
	sub	r27,r27
	movw	r30,r26
	rjmp	DivDD4

; ----- shift dividend HIGH left (on input: carry from dividend LOW)

DivDD2:	adc	r26,r26
	adc	r27,r27
	adc	r30,r30
	adc	r31,r31

; ----- compare dividend HIGH with divisor

	cp	r26,r18
	cpc	r27,r19
	cpc	r30,r20
	cpc	r31,r21
	brcs	DivDD4		; skip if dividend HIGH < divisor

; ----- subtract divisor from dividend HIGH (only in case if dividend HIGH >= divisor)

	sub	r26,r18
	sbc	r27,r19
	sbc	r30,r20
	sbc	r31,r21

; ----- shift dividend LOW left (add Carry on input if dividend HIGH < divisor)

DivDD4:	adc	r22,r22
	adc	r23,r23
	adc	r24,r24
	adc	r25,r25

; ----- next loop

	dec	r1
	brne	DivDD2

; ----- invert quotient bits

	com	r22
	com	r23
	com	r24
	com	r25

; ----- store remainder

	cp	r16,r1
	cpc	r17,r1
	breq	DivDD6

	movw	r0,r30
	movw	r30,r16
	st	Z,r26
	std	Z+1,r27
	std	Z+2,r0
	std	Z+3,r1
	eor	r1,r1

DivDD6:	ret

#endif

; ----------------------------------------------------------------------------
;                   Divide unsigned QWORD by QWORD, result is QWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22:R21:R20:R19:R18 = dividend (N7:N6:N5:N4:N3:N2:N1:N0)
;        R17:R16:R15:R14:R13:R12:R11:R10 = divisor (D7:D6:D5:D4:D3:D2:D1:D0)
;        R9:R8 = pointer to QWORD to store remainder (NULL=no remainder)
; OUTPUT: R25:R24:R23:R22:R21:R20:R19:R18 = quotient (Q7:Q6:Q5:Q4:Q3:Q2:Q1:Q0)
;         R1=0
; DESTROYS: R31, R30, R27, R26
; takes 126 us on 16 MHz, 148 bytes
; ----------------------------------------------------------------------------
; dividend HIGH: R31:R30:R29:R28:R27:R26:R7:R6

#ifdef USE_DIVQQ

.global DivQQ

; ----- push registers

DivQQ:	push	r29
	push	r28
	push	r7
	push	r6

; ----- number of loops

	ldi	r26,64+1
	mov	r1,r26

; ----- clear dividend HIGH

	sub	r30,r30
	sub	r31,r31
	movw	r28,r30
	movw	r26,r30
	movw	r6,r30
	rjmp	DivQQ4

; ----- shift dividend HIGH left (on input: carry from dividend LOW)

DivQQ2:	adc	r6,r6
	adc	r7,r7
	adc	r26,r26
	adc	r27,r27
	adc	r28,r28
	adc	r29,r29
	adc	r30,r30
	adc	r31,r31

; ----- compare dividend HIGH with divisor

	cp	r6,r10
	cpc	r7,r11
	cpc	r26,r12
	cpc	r27,r13
	cpc	r28,r14
	cpc	r29,r15
	cpc	r30,r16
	cpc	r31,r17
	brcs	DivQQ4		; skip if dividend HIGH < divisor

; ----- subtract divisor from dividend HIGH (only in case if dividend HIGH >= divisor)

	sub	r6,r10
	sbc	r7,r11
	sbc	r26,r12
	sbc	r27,r13
	sbc	r28,r14
	sbc	r29,r15
	sbc	r30,r16
	sbc	r31,r17

; ----- shift dividend LOW left (add Carry on input if dividend HIGH < divisor)

DivQQ4:	adc	r18,r18
	adc	r19,r19
	adc	r20,r20
	adc	r21,r21
	adc	r22,r22
	adc	r23,r23
	adc	r24,r24
	adc	r25,r25

; ----- next loop

	dec	r1
	brne	DivQQ2

; ----- invert quotient bits

	com	r18
	com	r19
	com	r20
	com	r21
	com	r22
	com	r23
	com	r24
	com	r25

; ----- store remainder

	cp	r8,r1
	cpc	r9,r1
	breq	DivQQ6

	movw	r0,r30
	movw	r30,r8
	st	Z,r6
	std	Z+1,r7
	std	Z+2,r26
	std	Z+3,r27
	std	Z+4,r28
	std	Z+5,r29
	std	Z+6,r0
	std	Z+7,r1
	eor	r1,r1

; ----- pop registers

DivQQ6:	pop	r6
	pop	r7
	pop	r28
	pop	r29
	ret

#endif
