; ****************************************************************************
;
;                                     Multiply
;
; ****************************************************************************

; 8-bit result
;#define USE_MULBB	; flag - use MulBB

; 16-bit result
;#define USE_MULWB	; flag - use MulWB
#define USE_MULWW	; flag - use MulWW
;#define USE_MULWBB	; flag - use MulWBB

; 32-bit result
;#define USE_MULDWB	; flag - use MulDWB
;#define USE_MULDWW	; flag - use MulDWW
;#define USE_MULDB	; flag - use MulDB
;#define USE_MULDW	; flag - use MulDW
;#define USE_MULDD	; flag - use MulDD

; 64-bit result
;#define USE_MULQDB	; flag - use MulQDB
;#define USE_MULQDW	; flag - use MulQDW
#define USE_MULQDD	; flag - use MulQDD
#define USE_MULQB	; flag - use MulQB
;#define USE_MULQW	; flag - use MulQW
;#define USE_MULQD	; flag - use MulQD
;#define USE_MULQQ	; flag - use MulQQ

	.text

; ----------------------------------------------------------------------------
;                   Multiply unsigned BYTE x BYTE, result is BYTE
; ----------------------------------------------------------------------------
; INPUT: R24 = 1st multiplier (N0)
;        R22 = 2nd multiplier (N1)
; OUTPUT: R24 = result (M0)
;         R1=0
; DESTROYS: R0
; takes 1.1 us on 16 MHz, 8 bytes
; ----------------------------------------------------------------------------

#ifdef USE_MULBB

.global MulBB

; ----- N0*N1 -> R24

MulBB:	mul	r22,r24
	mov	r24,r0

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned BYTE x BYTE, result is WORD
; ----------------------------------------------------------------------------
; INPUT: R24 = 1st multiplier (N0)
;        R22 = 2nd multiplier (N1)
; OUTPUT: R25:R24 = result (M1:M0)
;         R1=0
; DESTROYS: R0
; takes 1.2 us on 16 MHz, 8 bytes
; ----------------------------------------------------------------------------

#ifdef USE_MULWBB

.global MulWBB

; ----- N0*N1 -> R25:R24

MulWBB:
	mul	r22,r24
	movw	r24,r0

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned WORD x BYTE, result is WORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24 = 1st multiplier (N1:N0)
;        R22 = 2nd multiplier (N2)
; OUTPUT: R25:R24 = result (M1:M0)
;         R1=0
; DESTROYS: R19, R18, R0
; takes 1.6 us on 16 MHz, 14 bytes
; ----------------------------------------------------------------------------

#ifdef USE_MULWB

.global MulWB

; ----- save 1st multiplier to R19:R18

MulWB:	movw	r18,r24

; ----- N0*N2 (level M1:M0) -> R24:R25

	mul	r22,r18
	movw	r24,r0

; ----- N1*N2 (level M1) add to R25

	mul	r22,r19
	add	r25,r0

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned WORD x WORD, result is WORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24 = 1st multiplier (N1:N0)
;        R23:R22 = 2nd multiplier (N3:N2)
; OUTPUT: R25:R24 = result (M1:M0)
;         R1=0
; DESTROYS: R19, R18, R0
; takes 1.9 us on 16 MHz, 18 bytes, 15 clocks
; ----------------------------------------------------------------------------

#ifdef USE_MULWW

.global MulWW

; ----- save 1st multiplier to R19:R18

MulWW:	movw	r18,r24		; [1]

; ----- N0*N2 (level M1:M0) -> R24:R25

	mul	r22,r18		; [2]
	movw	r24,r0		; [1]

; ----- N1*N2 (level M1) add to R25

	mul	r22,r19		; [2]
	add	r25,r0		; [1]

; ----- N0*N3 (level M1) add to R25

	mul	r23,r18		; [2]
	add	r25,r0		; [1]

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1		; [1]
	ret			; [4,5]

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned WORD x BYTE, result is DWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24 = 1st multiplier (N1:N0)
;        R22 = 2nd multiplier (N2)
; OUTPUT: R25:R24:R23:R22 = result (M3:M2:M1:M0), R25 (M3) is always 0
;         R1=0
; DESTROYS: R26, R0
; takes 2.0 us on 16 MHz, 20 bytes
; ----------------------------------------------------------------------------

#ifdef USE_MULDWB

.global MulDWB

; ----- save 2nd multiplier to R26

MulDWB:	mov	r26,r22

; ----- N2*N0 (level M1:M0) -> R23:R22

	mul	r26,r24
	movw	r22,r0

; ----- N2*N1 (level M2:M1) add to R24:R23

	mul	r26,r25
	eor	r25,r25
	eor	r24,r24
	add	r23,r0
	adc	r24,r1

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned WORD x WORD, result is DWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24 = 1st multiplier (N1:N0)
;        R23:R22 = 2nd multiplier (N3:N2)
; OUTPUT: R25:R24:R23:R22 = result (M3:M2:M1:M0)
;         R1=0
; DESTROYS: R31, R30, R27, R26, R0
; takes 2.7 us on 16 MHz, 34 bytes
; ----------------------------------------------------------------------------

#ifdef USE_MULDWW

.global MulDWW

; ----- N0*N2 (level M1:M0) -> R27:R26

MulDWW:	mul	r24,r22
	movw	r26,r0

; ----- N1*N3 (level M3:M2) -> R31:R30

	mul	r25,r23
	movw	r30,r0

; ----- N2*N1 (level M2:M1) add to R31:R30:R27

	mul	r22,r25
	eor	r25,r25
	add	r27,r0
	adc	r30,r1 
	adc	r31,r25

; ----- N3*N0 (level M2:M1) add to R31:R30:R27

	mul	r23,r24
	add	r27,r0
	adc	r30,r1 
	adc	r31,r25

; ----- store result R31:R30:R27:R26 -> R25:R24:R23:R22

	movw	r22,r26
	movw	r24,r30

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned DWORD x BYTE, result is DWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22 = 1st multiplier (N3:N2:N1:N0)
;        R20 = 2nd multiplier (N4)
; OUTPUT: R25:R24:R23:R22 = result (M3:M2:M1:M0)
;         R1=0
; DESTROYS: R27, R26, R0
; takes 2.2 us on 16 MHz, 30 bytes
; ----------------------------------------------------------------------------

#ifdef USE_MULDB

.global MulDB

; ----- N2*N4 (level M3:M2) -> R27:R26

MulDB:	mul	r24,r20
	movw	r26,r0

; ----- N3*N4 (level M3) add to R27

	mul	r25,r20
	add	r27,r0

; ----- save HIGH result to R25:R24

	movw	r24,r26

; ----- set N1:N0 to R27:R26

	movw	r26,r22

; ----- N0:N4 (level M1:M0) -> R23:R22

	mul	r26,r20
	movw	r22,r0

; ----- R26 <- 0

	eor	r26,r26

; ----- N1*N4 (level M2:M1) add to R25:R24:R23

	mul	r27,r20
	add	r23,r0
	adc	r24,r1
	adc	r25,r26

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned DWORD x WORD, result is DWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22 = 1st multiplier (N3:N2:N1:N0)
;        R21:R20 = 2nd multiplier (N5:N4)
; OUTPUT: R25:R24:R23:R22 = result (M3:M2:M1:M0)
;         R1=0
; DESTROYS: R31, R30, R27, R26, R0
; takes 3.0 us on 16 MHz, 48 bytes
; ----------------------------------------------------------------------------

#ifdef USE_MULDW

.global MulDW

; ----- N0*N4 (level M1:M0) -> R27:R26

MulDW:	mul	r22,r20
	movw	r26,r0

; ----- N1*N5 (level M3:M2) -> R31:R30

	mul	r23,r21
	movw	r30,r0

; ----- N2*N4 (level M3:M2) add to R31:R30

	mul	r24,r20
	add	r30,r0
	adc	r31,r1 

; ----- N3*N4 (level M3) add to R31

	mul	r25,r20
	add	r31,r0

; ----- N2*N5 (level M3) add to R31

	mul	r24,r21
	add	r31,r0

; ----- R25 <- 0 (N3 no more needed)

	eor	r25,r25

; ----- N1*N4 (level M2:M1) add to R31:R30:R27

	mul	r23,r20
	add	r27,r0
	adc	r30,r1
	adc	r31,r25

; ----- N0*N5 (level M2:M1) add to R31:R30:R27

	mul	r22,r21
	add	r27,r0
	adc	r30,r1
	adc	r31,r25

; ----- store result R31:R30:R27:R26 -> R25:R24:R23:R22

	movw	r22,r26
	movw	r24,r30

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned DWORD x DWORD, result is DWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22 = 1st multiplier (N3:N2:N1:N0)
;        R21:R20:R19:R18 = 2nd multiplier (N7:N6:N5:N4)
; OUTPUT: R25:R24:R23:R22 = result (M3:M2:M1:M0)
;         R1=0
; DESTROYS: R31, R30, R27, R26, R0
; takes 3.6 us on 16 MHz, 62 bytes, 45 clocks
; ----------------------------------------------------------------------------

#ifdef USE_MULDD

.global MulDD

; ----- N0*N4 (level M1:M0) -> R27:R26 [3]

MulDD:	mul	r22,r18		; [2]
	movw	r26,r0		; [1]

; ----- N1*N5 (level M3:M2) -> R31:R30 [3]

	mul	r23,r19		; [2]
	movw	r30,r0		; [1]

; ----- N2*N4 (level M3:M2) add to R31:R30 [4]

	mul	r24,r18		; [2]
	add	r30,r0		; [1]
	adc	r31,r1 		; [1]

; ----- N0*N6 (level M3:M2) add to R31:R30 [4]

	mul	r22,r20		; [2]
	add	r30,r0		; [1]
	adc	r31,r1		; [1]

; ----- N3*N4 (level M3) add to R31 [3]

	mul	r25,r18		; [2]
	add	r31,r0		; [1]

; ----- N2*N5 (level M3) add to R31 [3]

	mul	r24,r19		; [2]
	add	r31,r0		; [1]

; ----- N1*N6 (level M3) add to R31 [3]

	mul	r23,r20		; [2]
	add	r31,r0		; [1]

; ----- N0*N7 (level M3) add to R31 [3]

	mul	r22,r21		; [2]
	add	r31,r0		; [1]

; ----- R25 <- 0 (N3 no more needed) [1]

	eor	r25,r25		; [1]

; ----- N1*N4 (level M2:M1) add to R31:R30:R27 [5]

	mul	r23,r18		; [2]
	add	r27,r0		; [1]
	adc	r30,r1		; [1]
	adc	r31,r25		; [1]

; ----- N0*N5 (level M2:M1) add to R31:R30:R27 [5]

	mul	r22,r19		; [2]
	add	r27,r0		; [1]
	adc	r30,r1		; [1]
	adc	r31,r25		; [1]

; ----- store result R31:R30:R27:R26 -> R25:R24:R23:R22 [2]

	movw	r22,r26		; [1]
	movw	r24,r30		; [1]

; ----- R1 <- 0 (set to standard value) [5,6]

	eor	r1,r1		; [1]
	ret			; [4,5]

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned DWORD x BYTE, result is QWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22 = 1st multiplier (N3:N2:N1:N0)
;        R20 = 2nd multiplier (N4)
; OUTPUT: R25:R24:R23:R22:R21:R20:R19:R18 = result (M7:M6:M5:M4:M3:M2:M1:M0)
;         R25, R24 and R23 are = 0
;         R1=0
; DESTROYS: R31, R30, R27, R26, R0
; takes 3.7 us on 16 MHz, 36 bytes
; ----------------------------------------------------------------------------

#ifdef USE_MULQDB

.global MulQDB

; ----- N4*odd (level M1) -> temporary R31:R30:R27:R26 (max. FE 01 FE 01 - 00)

MulQDB:	mul	r20,r23
	movw	r26,r0

	mul	r20,r25
	movw	r30,r0

; ----- N4*even (level M0) -> result R21:R20:R19:R18 (max. FE 01 FE 01)

	mul	r20,r22
	movw	r18,r0

	mul	r20,r24
	movw	r20,r0

; ----- add temporary to accumulator (max. FE FF FF FF 01)

	eor	r22,r22		; R22 <- 0
	add	r19,r26
	adc	r20,r27
	adc	r21,r30
	adc	r22,r31

; ----- clear unused high bytes

	eor	r25,r25
	eor	r24,r24
	eor	r23,r23

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned DWORD x WORD, result is QWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22 = 1st multiplier (N3:N2:N1:N0)
;        R21:R20 = 2nd multiplier (N5:N4)
; OUTPUT: R25:R24:R23:R22:R21:R20:R19:R18 = result (M7:M6:M5:M4:M3:M2:M1:M0)
;         R25 and R24 are = 0
;         R1=0
; DESTROYS: R31, R30, R27, R26, R0
; takes 5.6 us on 16 MHz, 76 bytes
; ----------------------------------------------------------------------------
; Accumulator: R31:R30:R9:R8:R19:R18
; Temporary: R27:R26, R1:R0

#ifdef USE_MULQDW

.global MulQDW

; ----- push registers

MulQDW: push	r8
	push	r9

; ----- N4*even (level M0) -> accumulator R9:R8:R19:R18 (max. FE 01 FE 01)

	mul	r20,r24
	movw	r8,r0

	mul	r20,r22
	movw	r18,r0

; ----- N4*odd (level M1) -> temporary R27:R26:R1:R0 (max. FE 01 FE 01 - 00)

	mul	r20,r25
	movw	r26,r0

	mul	r20,r23

; ----- add temporary to accumulator (max. FE FF FF FF 01)

	eor	r30,r30		; R30 <- 0
	add	r19,r0
	adc	r8,r1
	adc	r9,r26
	adc	r30,r27

; ----- N5*even (level M1) -> temporary R27:R26:R1:R0 (max. FE 01 FE 01 - 00)

	mul	r21,r24
	movw	r26,r0

	mul	r21,r22

; ----- add temporary to accumulator (max. 01 FD 01 FE 00 01)

	eor	r31,r31		; R31 <- 0
	add	r19,r0
	adc	r8,r1
	adc	r9,r26
	adc	r30,r27
	adc	r31,r31

; ----- N5*odd (level M2) -> temporary R27:R26:R1:R0 (max. FE 01 FE 01 - 00 00)

	mul	r21,r25
	movw	r26,r0

	mul	r21,r23

; ----- add temporary to accumulator (max. FF FE FF FF 00 01)

	add	r8,r0
	adc	r9,r1
	adc	r30,r26
	adc	r31,r27

; ----- store result

	movw	r20,r8
	movw	r22,r30
	eor	r24,r24
	eor	r25,r25

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1

; ----- pop registers

	pop	r9
	pop	r8
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned DWORD x DWORD, result is QWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22 = 1st multiplier (N3:N2:N1:N0)
;        R21:R20:R19:R18 = 2nd multiplier (N7:N6:N5:N4)
; OUTPUT: R25:R24:R23:R22:R21:R20:R19:R18 = result (M7:M6:M5:M4:M3:M2:M1:M0)
;         R1=0
; DESTROYS: R31, R30, R27, R26, R0
; takes 9.4 us on 16 MHz, 156 bytes
; ----------------------------------------------------------------------------
; Accumulator: R31:R30:R9:R8:R7:R6:R5:R4
; Temporary: R27:R26, R1:R0

#ifdef USE_MULQDD

.global MulQDD

; ----- push registers

MulQDD: push	r4
	push	r5
	push	r6
	push	r7
	push	r8
	push	r9

; ----- N4*even (level M0) -> accumulator R7:R6:R5:R4 (max. FE 01 FE 01)

	mul	r18,r24
	movw	r6,r0

	mul	r18,r22
	movw	r4,r0

; ----- N4*odd (level M1) -> temporary R27:R26:R1:R0 (max. FE 01 FE 01 - 00)

	mul	r18,r25
	movw	r26,r0

	mul	r18,r23

; ----- add temporary to accumulator (max. FE FF FF FF 01)

	eor	r8,r8		; R8 <- 0
	add	r5,r0
	adc	r6,r1
	adc	r7,r26
	adc	r8,r27

; ----- N5*even (level M1) -> temporary R27:R26:R1:R0 (max. FE 01 FE 01 - 00)

	mul	r19,r24
	movw	r26,r0

	mul	r19,r22

; ----- add temporary to accumulator (max. 01 FD 01 FE 00 01)

	eor	r9,r9		; R9 <- 0
	add	r5,r0
	adc	r6,r1
	adc	r7,r26
	adc	r8,r27
	adc	r9,r9

; ----- N5*odd (level M2) -> temporary R27:R26:R1:R0 (max. FE 01 FE 01 - 00 00)

	mul	r19,r25
	movw	r26,r0

	mul	r19,r23

; ----- add temporary to accumulator (max. FF FE FF FF 00 01)

	add	r6,r0
	adc	r7,r1
	adc	r8,r26
	adc	r9,r27

; ----- N6*even (level M2) -> temporary R27:R26:R1:R0 (max. FE 01 FE 01 - 00 00)

	mul	r20,r24
	movw	r26,r0

	mul	r20,r22

; ----- add temporary to accumulator (max. 01 FE 00 FE 00 00 01)

	eor	r30,r30		; R30 <- 0
	add	r6,r0
	adc	r7,r1
	adc	r8,r26
	adc	r9,r27
	adc	r30,r30

; ----- N6*odd (level M3) -> temporary R27:R26:R1:R0 (max. FE 01 FE 01 - 00 00 00)

	mul	r20,r25
	movw	r26,r0

	mul	r20,r23

; ----- add temporary to accumulator (max. FF FF FE FF 00 00 01)

	add	r7,r0
	adc	r8,r1
	adc	r9,r26
	adc	r30,r27

; ----- N7*even (level M3) -> temporary R27:R26:R1:R0 (max. FE 01 FE 01 - 00 00 00)

	mul	r21,r24
	movw	r26,r0

	mul	r21,r22

; ----- add temporary to accumulator (max. 01 FE 01 FD 00 00 00 01)

	eor	r31,r31		; R31 <- 0
	add	r7,r0
	adc	r8,r1
	adc	r9,r26
	adc	r30,r27
	adc	r31,r31

; ----- N7*odd (level M4) -> temporary R27:R26:R1:R0 (max. FE 01 FE 01 - 00 00 00 00)

	mul	r21,r25
	movw	r26,r0

	mul	r21,r23

; ----- add temporary to accumulator (max. FF FF FF FE 00 00 00 01)

	add	r8,r0
	adc	r9,r1
	adc	r30,r26
	adc	r31,r27

; ----- store result

	movw	r18,r4
	movw	r20,r6
	movw	r22,r8
	movw	r24,r30

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1

; ----- pop registers

	pop	r9
	pop	r8
	pop	r7
	pop	r6
	pop	r5
	pop	r4
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned QWORD x BYTE, result is QWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22:R21:R20:R19:R18 = 1st multiplier (N7:N6:N5:N4:N3:N2:N1:N0)
;        R16 = 2nd multiplier (N8)
; OUTPUT: R25:R24:R23:R22:R21:R20:R19:R18 = result (M7:M6:M5:M4:M3:M2:M1:M0)
;         R1=0
; DESTROYS: R31, R30, R0
; takes 6 us on 16 MHz, 78 bytes
; ----------------------------------------------------------------------------

#ifdef USE_MULQB

.global MulQB

; ----- R31 <- 0

MulQB:	eor	r31,r31

; ----- N8*N0 (level M0)

	mul	r16,r18
	mov	r18,r0
	mov	r30,r1

; ----- N8*N1 (level M1)

	mul	r16,r19
	add	r0,r30
	adc	r1,r31
	mov	r19,r0
	mov	r30,r1

; ----- N8*N2 (level M2)

	mul	r16,r20
	add	r0,r30
	adc	r1,r31
	mov	r20,r0
	mov	r30,r1

; ----- N8*N3 (level M3)

	mul	r16,r21
	add	r0,r30
	adc	r1,r31
	mov	r21,r0
	mov	r30,r1

; ----- N8*N4 (level M4)

	mul	r16,r22
	add	r0,r30
	adc	r1,r31
	mov	r22,r0
	mov	r30,r1

; ----- N8*N5 (level M5)

	mul	r16,r23
	add	r0,r30
	adc	r1,r31
	mov	r23,r0
	mov	r30,r1

; ----- N8*N6 (level M6)

	mul	r16,r24
	add	r0,r30
	adc	r1,r31
	mov	r24,r0
	mov	r30,r1

; ----- N8*N7 (level M7)

	mul	r16,r25
	add	r0,r30
	mov	r25,r0

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned QWORD x WORD, result is QWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22:R21:R20:R19:R18 = 1st multiplier (N7:N6:N5:N4:N3:N2:N1:N0)
;        R17:R16 = 2nd multiplier (N9:N8)
; OUTPUT: R25:R24:R23:R22:R21:R20:R19:R18 = result (M7:M6:M5:M4:M3:M2:M1:M0)
;         R1=0
; DESTROYS: R31, R30, R27, R26, R0
; takes 10 us on 16 MHz, 142 bytes
; ----------------------------------------------------------------------------
; Accumulator: R9:R8:R7:R6:R5:R4:R3:R2
; Temporary: R29:R31:R30:R27:R26, R1:R0

#ifdef USE_MULQW

.global MulQW

; ----- push registers

MulQW:  push	r2
	push	r3
	push	r4
	push	r5
	push	r6
	push	r7
	push	r8
	push	r9
	push	r29

; ----- N8*even (level M0) -> accumulator

	mul	r16,r24
	movw	r8,r0

	mul	r16,r22
	movw	r6,r0

	mul	r16,r20
	movw	r4,r0

	mul	r16,r18
	movw	r2,r0

; ----- N8*odd (level M1) -> temporary

	mul	r16,r25
	mov	r29,r0

	mul	r16,r23
	movw	r30,r0

	mul	r16,r21
	movw	r26,r0

	mul	r16,r19

; ----- add temporary to accumulator

	add	r3,r0
	adc	r4,r1
	adc	r5,r26
	adc	r6,r27
	adc	r7,r30
	adc	r8,r31
	adc	r9,r29

; ----- N9*even (level M1) -> temporary

	mul	r17,r24
	mov	r29,r0

	mul	r17,r22
	movw	r30,r0

	mul	r17,r20
	movw	r26,r0

	mul	r17,r18

; ----- add temporary to accumulator

	add	r3,r0
	adc	r4,r1
	adc	r5,r26
	adc	r6,r27
	adc	r7,r30
	adc	r8,r31
	adc	r9,r29

; ----- N9*odd (level M2) -> temporary

	mul	r17,r23
	movw	r30,r0

	mul	r17,r21
	movw	r26,r0

	mul	r17,r19

; ----- add temporary to accumulator

	add	r4,r0
	adc	r5,r1
	adc	r6,r26
	adc	r7,r27
	adc	r8,r30
	adc	r9,r31

; ----- store result

	movw	r18,r2
	movw	r20,r4
	movw	r22,r6
	movw	r24,r8

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1

; ----- pop registers

	pop	r29
	pop	r9
	pop	r8
	pop	r7
	pop	r6
	pop	r5
	pop	r4
	pop	r3
	pop	r2
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned QWORD x DWORD, result is QWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22:R21:R20:R19:R18 = 1st multiplier (N7:N6:N5:N4:N3:N2:N1:N0)
;        R17:R16:R15:R14 = 2nd multiplier (N11:N10:N9:N8)
; OUTPUT: R25:R24:R23:R22:R21:R20:R19:R18 = result (M7:M6:M5:M4:M3:M2:M1:M0)
;         R1=0
; DESTROYS: R31, R30, R27, R26, R0
; takes 13 us on 16 MHz, 218 bytes
; ----------------------------------------------------------------------------
; Accumulator: R9:R8:R7:R6:R5:R4:R3:R2
; Temporary: R29:R31:R30:R27:R26, R1:R0

#ifdef USE_MULQD

.global MulQD

; ----- push registers

MulQD:  push	r2
	push	r3
	push	r4
	push	r5
	push	r6
	push	r7
	push	r8
	push	r9
	push	r29

; ----- N8*even (level M0) -> accumulator

	mul	r14,r24
	movw	r8,r0

	mul	r14,r22
	movw	r6,r0

	mul	r14,r20
	movw	r4,r0

	mul	r14,r18
	movw	r2,r0

; ----- N8*odd (level M1) -> temporary

	mul	r14,r25
	mov	r29,r0

	mul	r14,r23
	movw	r30,r0

	mul	r14,r21
	movw	r26,r0

	mul	r14,r19

; ----- add temporary to accumulator

	add	r3,r0
	adc	r4,r1
	adc	r5,r26
	adc	r6,r27
	adc	r7,r30
	adc	r8,r31
	adc	r9,r29

; ----- N9*even (level M1) -> temporary

	mul	r15,r24
	mov	r29,r0

	mul	r15,r22
	movw	r30,r0

	mul	r15,r20
	movw	r26,r0

	mul	r15,r18

; ----- add temporary to accumulator

	add	r3,r0
	adc	r4,r1
	adc	r5,r26
	adc	r6,r27
	adc	r7,r30
	adc	r8,r31
	adc	r9,r29

; ----- N9*odd (level M2) -> temporary

	mul	r15,r23
	movw	r30,r0

	mul	r15,r21
	movw	r26,r0

	mul	r15,r19

; ----- add temporary to accumulator

	add	r4,r0
	adc	r5,r1
	adc	r6,r26
	adc	r7,r27
	adc	r8,r30
	adc	r9,r31

; ----- N10*even (level M2) -> temporary

	mul	r16,r22
	movw	r30,r0

	mul	r16,r20
	movw	r26,r0

	mul	r16,r18

; ----- add temporary to accumulator

	add	r4,r0
	adc	r5,r1
	adc	r6,r26
	adc	r7,r27
	adc	r8,r30
	adc	r9,r31

; ----- N10*odd (level M3) -> temporary

	mul	r16,r23
	mov	r30,r0

	mul	r16,r21
	movw	r26,r0

	mul	r16,r19

; ----- add temporary to accumulator

	add	r5,r0
	adc	r6,r1
	adc	r7,r26
	adc	r8,r27
	adc	r9,r30

; ----- N11*even (level M3) -> temporary

	mul	r17,r22
	mov	r30,r0

	mul	r17,r20
	movw	r26,r0

	mul	r17,r18

; ----- add temporary to accumulator

	add	r5,r0
	adc	r6,r1
	adc	r7,r26
	adc	r8,r27
	adc	r9,r30

; ----- N11*odd (level M4) -> temporary

	mul	r17,r21
	movw	r26,r0

	mul	r17,r19

; ----- add temporary to accumulator

	add	r6,r0
	adc	r7,r1
	adc	r8,r26
	adc	r9,r27

; ----- store result

	movw	r18,r2
	movw	r20,r4
	movw	r22,r6
	movw	r24,r8

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1

; ----- pop registers

	pop	r29
	pop	r9
	pop	r8
	pop	r7
	pop	r6
	pop	r5
	pop	r4
	pop	r3
	pop	r2
	ret

#endif

; ----------------------------------------------------------------------------
;                   Multiply unsigned QWORD x QWORD, result is QWORD
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22:R21:R20:R19:R18 = 1st multiplier (N7:N6:N5:N4:N3:N2:N1:N0)
;        R17:R16:R15:R14:R13:R12:R11:R10 = 2nd multiplier (N15:N14:N13:N12:N11:N10:N9:N8)
; OUTPUT: R25:R24:R23:R22:R21:R20:R19:R18 = result (M7:M6:M5:M4:M3:M2:M1:M0)
;         R1=0
; DESTROYS: R31, R30, R27, R26, R0
; takes 18 us on 16 MHz, 276 bytes
; ----------------------------------------------------------------------------
; Accumulator: R9:R8:R7:R6:R5:R4:R3:R2
; Temporary: R29:R31:R30:R27:R26, R1:R0

#ifdef USE_MULQQ

.global MulQQ

; ----- push registers

MulQQ:  push	r2
	push	r3
	push	r4
	push	r5
	push	r6
	push	r7
	push	r8
	push	r9
	push	r29

; ----- N0*even (level M0) -> accumulator

	mul	r18,r16
	movw	r8,r0

	mul	r18,r14
	movw	r6,r0

	mul	r18,r12
	movw	r4,r0

	mul	r18,r10
	movw	r2,r0

; ----- N0*odd (level M1) -> temporary

	mul	r18,r17
	mov	r29,r0

	mul	r18,r15
	movw	r30,r0

	mul	r18,r13
	movw	r26,r0

	mul	r18,r11

; ----- add temporary to accumulator

	add	r3,r0
	adc	r4,r1
	adc	r5,r26
	adc	r6,r27
	adc	r7,r30
	adc	r8,r31
	adc	r9,r29

; ----- N1*even (level M1) -> temporary

	mul	r19,r16
	mov	r29,r0

	mul	r19,r14
	movw	r30,r0

	mul	r19,r12
	movw	r26,r0

	mul	r19,r10

; ----- add temporary to accumulator

	add	r3,r0
	adc	r4,r1
	adc	r5,r26
	adc	r6,r27
	adc	r7,r30
	adc	r8,r31
	adc	r9,r29

; ----- N1*odd (level M2) -> temporary

	mul	r19,r15
	movw	r30,r0

	mul	r19,r13
	movw	r26,r0

	mul	r19,r11

; ----- add temporary to accumulator

	add	r4,r0
	adc	r5,r1
	adc	r6,r26
	adc	r7,r27
	adc	r8,r30
	adc	r9,r31

; ----- N2*even (level M2) -> temporary

	mul	r20,r14
	movw	r30,r0

	mul	r20,r12
	movw	r26,r0

	mul	r20,r10

; ----- add temporary to accumulator

	add	r4,r0
	adc	r5,r1
	adc	r6,r26
	adc	r7,r27
	adc	r8,r30
	adc	r9,r31

; ----- N2*odd (level M3) -> temporary

	mul	r20,r15
	mov	r30,r0

	mul	r20,r13
	movw	r26,r0

	mul	r20,r11

; ----- add temporary to accumulator

	add	r5,r0
	adc	r6,r1
	adc	r7,r26
	adc	r8,r27
	adc	r9,r30

; ----- N3*even (level M3) -> temporary

	mul	r21,r14
	mov	r30,r0

	mul	r21,r12
	movw	r26,r0

	mul	r21,r10

; ----- add temporary to accumulator

	add	r5,r0
	adc	r6,r1
	adc	r7,r26
	adc	r8,r27
	adc	r9,r30

; ----- N3*odd (level M4) -> temporary

	mul	r21,r13
	movw	r26,r0

	mul	r21,r11

; ----- add temporary to accumulator

	add	r6,r0
	adc	r7,r1
	adc	r8,r26
	adc	r9,r27

; ----- N4*even (level M4) -> temporary

	mul	r22,r12
	movw	r26,r0

	mul	r22,r10

; ----- add temporary to accumulator

	add	r6,r0
	adc	r7,r1
	adc	r8,r26
	adc	r9,r27

; ----- N4*odd (level M5) -> temporary

	mul	r22,r13
	mov	r26,r0

	mul	r22,r11

; ----- add temporary to accumulator

	add	r7,r0
	adc	r8,r1
	adc	r9,r26

; ----- N5*even (level M5) -> temporary

	mul	r23,r12
	mov	r26,r0

	mul	r23,r10

; ----- add temporary to accumulator

	add	r7,r0
	adc	r8,r1
	adc	r9,r26

; ----- N5*odd (level M6) -> temporary

	mul	r23,r11

; ----- add temporary to accumulator

	add	r8,r0
	adc	r9,r1

; ----- N6*even (level M6) -> temporary

	mul	r24,r10

; ----- add temporary to accumulator

	add	r8,r0
	adc	r9,r1

; ----- N6*odd (level M7) -> temporary

	mul	r24,r11

; ----- add temporary to accumulator

	add	r9,r0

; ----- N7*even (level M7) -> temporary

	mul	r25,r10

; ----- add temporary to accumulator

	add	r9,r0

; ----- store result

	movw	r18,r2
	movw	r20,r4
	movw	r22,r6
	movw	r24,r8

; ----- R1 <- 0 (set to standard value)

	eor	r1,r1

; ----- pop registers

	pop	r29
	pop	r9
	pop	r8
	pop	r7
	pop	r6
	pop	r5
	pop	r4
	pop	r3
	pop	r2
	ret

#endif
