; ****************************************************************************
;
;                                 Shifts
;
; ****************************************************************************

; Left
;#define USE_SHLB	; flag - use ShlB
;#define USE_SHLW	; flag - use ShlW
;#define USE_SHLD	; flag - use ShlD
;#define USE_SHLQ	; flag - use ShlQ

; Right
;#define USE_SHRB	; flag - use ShrB
;#define USE_SHRW	; flag - use ShrW
;#define USE_SHRD	; flag - use ShrD
#define USE_SHRQ	; flag - use ShrQ

	.text

; ----------------------------------------------------------------------------
;                                 Shift BYTE left
; ----------------------------------------------------------------------------
; INPUT: R24 = operand
;        R22 = shifts (signed)
; OUTPUT: R24 = result
; DESTROYS: R22
; takes average 2.8 us on 16 MHz, 8 bytes
; ----------------------------------------------------------------------------

#ifdef USE_SHLB

.global ShlB

ShlB2:	add	r24,r24
ShlB:	dec	r22
	brpl	ShlB2
	ret

#endif

; ----------------------------------------------------------------------------
;                                 Shift WORD left
; ----------------------------------------------------------------------------
; INPUT: R25:R24 = operand
;        R22 = shifts (signed)
; OUTPUT: R25:R24 = result
; DESTROYS: R22
; takes average 4.5 us on 16 MHz, 10 bytes
; ----------------------------------------------------------------------------

#ifdef USE_SHLW

.global ShlW

ShlW2:	add	r24,r24
	adc	r25,r25
ShlW:	dec	r22
	brpl	ShlW2
	ret

#endif

; ----------------------------------------------------------------------------
;                                 Shift DWORD left
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22 = operand
;        R20 = shifts (signed)
; OUTPUT: R25:R24:R23:R22 = result
; DESTROYS: R20
; takes average 9.3 us on 16 MHz, 14 bytes
; ----------------------------------------------------------------------------

#ifdef USE_SHLD

.global ShlD

ShlD2:	add	r22,r22
	adc	r23,r23
	adc	r24,r24
	adc	r25,r25
ShlD:	dec	r20
	brpl	ShlD2
	ret

#endif

; ----------------------------------------------------------------------------
;                                 Shift QWORD left
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22:R21:R20:R19:R18 = operand
;        R16 = shifts (signed)
; OUTPUT: R25:R24:R23:R22:R21:R20:R19:R18 = result
; DESTROYS: R31
; takes average 25 us on 16 MHz, 26 bytes
; ----------------------------------------------------------------------------

#ifdef USE_SHLQ

.global ShlQ

ShlQ:	mov	r31,r16
	rjmp	ShlQ4

ShlQ2:	add	r18,r18
	adc	r19,r19
	adc	r20,r20
	adc	r21,r21
	adc	r22,r22
	adc	r23,r23
	adc	r24,r24
	adc	r25,r25

ShlQ4:	dec	r31
	brpl	ShlQ2
	ret

#endif

; ----------------------------------------------------------------------------
;                                 Shift BYTE right
; ----------------------------------------------------------------------------
; INPUT: R24 = operand
;        R22 = shifts (signed)
; OUTPUT: R24 = result
; DESTROYS: R22
; takes average 2.8 us on 16 MHz, 8 bytes
; ----------------------------------------------------------------------------

#ifdef USE_SHRB

.global ShrB

ShrB2:	lsr	r24
ShrB:	dec	r22
	brpl	ShrB2
	ret

#endif

; ----------------------------------------------------------------------------
;                                 Shift WORD right
; ----------------------------------------------------------------------------
; INPUT: R25:R24 = operand
;        R22 = shifts (signed)
; OUTPUT: R25:R24 = result
; DESTROYS: R22
; takes average 4.5 us on 16 MHz, 10 bytes
; ----------------------------------------------------------------------------

#ifdef USE_SHRW

.global ShrW

ShrW2:	lsr	r25
	ror	r24
ShrW:	dec	r22
	brpl	ShrW2
	ret

#endif

; ----------------------------------------------------------------------------
;                                 Shift DWORD right
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22 = operand
;        R20 = shifts (signed)
; OUTPUT: R25:R24:R23:R22 = result
; DESTROYS: R20
; takes average 9.3 us on 16 MHz, 14 bytes
; ----------------------------------------------------------------------------

#ifdef USE_SHRD

.global ShrD

ShrD2:	lsr	r25
	ror	r24
	ror	r23
	ror	r22
ShrD:	dec	r20
	brpl	ShrD2
	ret

#endif

; ----------------------------------------------------------------------------
;                                 Shift QWORD right
; ----------------------------------------------------------------------------
; INPUT: R25:R24:R23:R22:R21:R20:R19:R18 = operand
;        R16 = shifts (signed)
; OUTPUT: R25:R24:R23:R22:R21:R20:R19:R18 = result
; DESTROYS: R31
; takes average 25 us on 16 MHz, 26 bytes
; ----------------------------------------------------------------------------

#ifdef USE_SHRQ

.global ShrQ

ShrQ:	mov	r31,r16
	rjmp	ShrQ4

ShrQ2:	lsr	r25
	ror	r24
	ror	r23
	ror	r22
	ror	r21
	ror	r20
	ror	r19
	ror	r18

ShrQ4:	dec	r31
	brpl	ShrQ2
	ret

#endif
