
// ****************************************************************************
//                                 
//                               Definitions
//
// ****************************************************************************

#define DIGITS	7	// number of digits
#define VALMAX	9999999UL // max. value (= DIGITS x '9')
#define VALMIN	1000000UL // min. value (= '1' + (DIGITS-1) x '0')

// ----------------------------------------------------------------------------
//                                  Constants
// ----------------------------------------------------------------------------

#define	B0	0x1
#define	B1	0x2
#define	B2	0x4
#define	B3	0x8
#define	B4	0x10
#define	B5	0x20
#define	B6	0x40
#define	B7	0x80
#define	B8	0x100
#define	B9	0x200
#define	B10	0x400
#define	B11	0x800
#define	B12	0x1000
#define	B13	0x2000
#define	B14	0x4000
#define	B15	0x8000
#define B16	0x10000
#define B17	0x20000
#define B18	0x40000
#define	B19	0x80000
#define B20	0x100000
#define B21	0x200000
#define B22	0x400000
#define B23	0x800000
#define B24	0x1000000
#define B25	0x2000000
#define B26	0x4000000
#define B27	0x8000000
#define B28	0x10000000
#define B29	0x20000000
#define B30	0x40000000
#define B31	0x80000000

#define BIT(pos) (1 << (pos))

// ----------------------------------------------------------------------------
//                                      data types
// ----------------------------------------------------------------------------

typedef signed char s8;
typedef unsigned char u8;
typedef signed short s16;
typedef unsigned short u16;
typedef signed long int s32;
typedef unsigned long int u32;
typedef signed long long int s64;
typedef unsigned long long int u64;

typedef unsigned char Bool;
#define True 1
#define False 0
#define NULL (void*)0

// ----------------------------------------------------------------------------
//                             extern functions
// ----------------------------------------------------------------------------

// ----- wait.S

// delay 1 us
extern void wait1us();

// delay 2..65537 us (enter 2..65536, or 0=65536 us, or 1=65537 us)
extern void waitus(u16 us);

// delay 1..65536 ms (0=65536 ms)
extern void waitms(u16 ms);

// ----- mul.S

// multiply unsigned BYTE x BYTE, result is BYTE
extern u8 MulBB(u8 num1, u8 num2);

// multiply unsigned BYTE x BYTE, result is WORD
extern u16 MulWBB(u8 num1, u8 num2);

// multiply unsigned WORD x BYTE, result is WORD
extern u16 MulWB(u16 num1, u8 num2);

// multiply unsigned WORD x WORD, result is WORD
extern u16 MulWW(u16 num1, u16 num2);

// multiply unsigned WORD x BYTE, result is DWORD (highest byte is always 0)
extern u32 MulDWB(u16 num1, u8 num2);

// multiply unsigned WORD x WORD, result is DWORD
extern u32 MulDWW(u16 num1, u16 num2);

// multiply unsigned DWORD x BYTE, result is DWORD
extern u32 MulDB(u32 num1, u8 num2);

// multiply unsigned DWORD x WORD, result is DWORD
extern u32 MulDW(u32 num1, u16 num2);

// multiply unsigned DWORD x DWORD, result is DWORD
extern u32 MulDD(u32 num1, u32 num2);

// multiply unsigned DWORD x BYTE, result is QWORD
extern u64 MulQDB(u32 num1, u8 num2);

// multiply unsigned DWORD x WORD, result is QWORD
extern u64 MulQDW(u32 num1, u16 num2);

// multiply unsigned DWORD x DWORD, result is QWORD
extern u64 MulQDD(u32 num1, u32 num2);

// multiply unsigned QWORD x BYTE, result is QWORD
extern u64 MulQB(u64 num1, u8 num2);

// multiply unsigned QWORD x WORD, result is QWORD
extern u64 MulQW(u64 num1, u16 num2);

// multiply unsigned QWORD x DWORD, result is QWORD
extern u64 MulQD(u64 num1, u32 num2);

// multiply unsigned QWORD x QWORD, result is QWORD
extern u64 MulQQ(u64 num1, u64 num2);

// ----- shift.S

// shift left
extern u8 ShlB(u8 num, s8 bit);
extern u16 ShlW(u16 num, s8 bit);
extern u32 ShlD(u32 num, s8 bit);
extern u64 ShlQ(u64 num, s8 bit);

// shift right
extern u8 ShrB(u8 num, s8 bit);
extern u16 ShrW(u16 num, s8 bit);
extern u32 ShrD(u32 num, s8 bit);
extern u64 ShrQ(u64 num, s8 bit);

// ----- div.S

extern u8 DivBB(u8 num, u8 d, u8* rem);
extern u16 DivWW(u16 num, u16 d, u16* rem);
extern u32 DivDD(u32 num, u32 d, u32* rem);
extern u64 DivQQ(u64 num, u64 d, u64* rem);

// ----- add.S

extern u64 AddQB(u64 num1, u8 num2);
extern u64 AddQW(u64 num1, u16 num2);
extern u64 AddQD(u64 num1, u32 num2);
extern u64 AddQQ(u64 num1, u64 num2);

// ----------------------------------------------------------------------------
//                                      includes
// ----------------------------------------------------------------------------

void main() __attribute__ ((noreturn));

// Generator modes (8 modes)
enum {
	GMODE_DEC = 0,		// A: decimal square
	GMODE_INT = 1,		// B: integer square
	GMODE_NOTE = 2,		// C: note square
	GMODE_SINE = 3,		// D: note sine
	GMODE_TRI = 4,		// E: note triangle
	GMODE_SAW = 5,		// F: note sawtooth
	GMODE_NOISE = 6,	// G: noise

	GMODE_NUM
};

// timer modes
enum {
	TMODE_FAST = 0,		// fast "fast PWM", ON at middle of period, OFF at end of period
	TMODE_SLOW,		// slow "fast PWM", flip-flop at end of period
	TMODE_CNT,		// counter mode, flip-flop after count of interrupts
	TMODE_PWM,		// PWM mode

	TMODE_NUM
};

// fast flag
#define FASTFLAG B3

// prescaler index mask
#define PRESMASK 0x7

#include <avr/io.h>	// ports
#include <avr/interrupt.h> // interrupts
#include <avr/pgmspace.h> // read const data
#include <avr/eeprom.h> // eeprom
#include "lcd.h"	// LCD display
#include "display.h"	// display
#include "main.h"	// main
#include "Timer1.h"	// Timer1
#include "Timer2.h"	// Timer2
#include "Waves.h"	// waves
#include "Button.h"	// buttons
