
// ****************************************************************************
//                                 
//                                Main function
//
// ****************************************************************************

#include "include.h"

// text - no signal
//const u8 TextF[] PROGMEM = "F = ";

// note names
const u8 NoteName[12] PROGMEM = { "CcDdEFfGgAaB" };

const u8 TextBrown[] PROGMEM = { "Brown Noise" };
const u8 TextPink[] PROGMEM =  { "Pink Noise" };
const u8 TextWhite[] PROGMEM = { "White Noise" };
const u8* TextNoise[3] = {  TextBrown, TextPink, TextWhite };

// display frequency
void DispFreq(u32 d)
{
	// frequency * 1e10
	u64 f = F_CPU*10000000000ULL;
	s8 scale = -10;

	// divide frequency by divider
	f = DivQQ(AddQD(f, d/2), d, NULL);

	// normalize result DOWN, to fit into DWORD
	while (f > 4000000000UL)
	{
		f = DivQQ(AddQB(f, 5), 10, NULL);
		scale++;
	}

	// display frequency
	DispUVal(f, scale, CHAR_HZ);
}

// display one generator
void DispGen(u8 row, u8 gen, u16 sel, u32 dv)
{
	u8 k;
	DispSetPos(row, 0);
	DispChar(gen+'A');
	DispChar(':');

	if ((gen >= GMODE_NOTE) && (gen <= GMODE_SAW))
	{
		k = sel % 12;
		DispChar(pgm_read_byte(&NoteName[k]));
		DispInt((s16)sel/12 - 4, False);
	}
	else
		DispUInt(sel+1);

	if (gen == GMODE_NOISE)
	{
		DispTab(4);
		DispTextROM(TextNoise[sel]);
	}
	else
	{
		DispTab(6);

		if ((gen > GMODE_NOTE) && (gen <= GMODE_SAW))
		{
			// frequency * 100 * coefficient
			u64 f = MulQDD(F_CPU*100, pgm_read_dword(&ToneSineInc[sel]));
			s8 scale = -2;

			// divide frequency by 256
			f = ShrQ(f, 8);

			// normalize result UP
			while (f < 1000000000000000000ULL)
			{
				f = MulQB(f, 10);
				scale--;
			}

			// display frequency
			DispUVal((u32)ShrQ(f, 32), scale, CHAR_HZ);
		}
		else
			DispFreq(dv);
	}

	DispClearRestLine();
}

// main function
void main()
{
	u8 n;

	// disable analog comparator (to save energy)
	ACSR |= BIT(ACD);

	// initialize LCD display
	LCD_Init();

	// enable global interrupt
	sei();

	// set Timer1 generator
	Timer1SetGen(GMODE_DEC);

	// set Timer2 generator
	Timer2SetGen(GMODE_DEC);

	// initialize buttons
	ButtonInit();

	while (True)
	{
		// G1SEL
		if (BtnCheck(BTN_G1SEL)) Timer1SetGen(Timer1Gen+1);

		// G1DN
		if (BtnCheck(BTN_G1DN)) Timer1SetSel(Timer1Sel[Timer1Gen]-1);
	
		// G1UP
		if (BtnCheck(BTN_G1UP)) Timer1SetSel(Timer1Sel[Timer1Gen]+1);

		// G2SEL
		if (BtnCheck(BTN_G2SEL)) Timer2SetGen(Timer2Gen+1);

		// G2DN
		if (BtnCheck(BTN_G2DN)) Timer2SetSel(Timer2Sel[Timer2Gen]-1);
	
		// G2UP
		if (BtnCheck(BTN_G2UP)) Timer2SetSel(Timer2Sel[Timer2Gen]+1);

		// display G1
		DispGen(0, Timer1Gen, Timer1SelNow, Timer1Div);

		// display G2
		DispGen(1, Timer2Gen, Timer2SelNow, Timer2Div);

		// wait 100 ms
		n = 100;
		if (Timer1Gen >= GMODE_SINE) n = 75;
		if (Timer2Gen >= GMODE_SINE) n = (u16)n*3/4;
		waitms(n);
	}
}
